/**
 *
 * Agora Real Time Engagement
 * Copyright (c) 2024 Agora IO. All rights reserved.
 *
 */

#import <Foundation/Foundation.h>

#import "AgoraRteEnumerates.h"

#if TARGET_OS_IPHONE
#import <UIKit/UIKit.h>
typedef UIView AgoraRteView;
#elif TARGET_OS_MAC
#import <AppKit/AppKit.h>
typedef NSView AgoraRteView;
#endif

@class AgoraRteError;
@class AgoraRte;

/**
 * The Rect class is used to manage the rect.
 * @since v4.4.0
 * @technical preview
 */
__attribute__((visibility("default"))) @interface AgoraRteRect : NSObject
- (instancetype _Nonnull)init;

- (void)setX:(int32_t)x;
- (int32_t)x;

- (void)setY:(int32_t)y;
- (int32_t)y;

- (void)setWidth:(int32_t)width;
- (int32_t)width;

- (void)setHeight:(int32_t)height;
- (int32_t)height;
@end

/**
 * The ViewConfig class is used to manage the view config.
 * @since v4.4.0
 * @technical preview
 */
__attribute__((visibility("default"))) @interface AgoraRteViewConfig : NSObject
- (instancetype _Nonnull)init;

- (void)setCropArea:(AgoraRteRect* _Nullable)cropArea error:(AgoraRteError* _Nullable)error;

- (AgoraRteRect* _Nullable)cropArea:(AgoraRteError* _Nullable)error;
@end

/**
 * The CanvasInitialConfig class is used to initialize the Canvas object.
 * @since v4.4.0
 * @technical preview
 */
__attribute__((visibility("default"))) @interface AgoraRteCanvasInitialConfig : NSObject
- (instancetype _Nonnull)init;
@end


/**
 * The CanvasConfig class is used to manage the canvas config.
 * @since v4.4.0
 * @technical preview
 */
__attribute__((visibility("default"))) @interface AgoraRteCanvasConfig : NSObject
- (instancetype _Nonnull)init;

- (void)setVideoRenderMode:(AgoraRteVideoRenderMode)mode error:(AgoraRteError * _Nullable)error;

- (AgoraRteVideoRenderMode)videoRenderMode:(AgoraRteError * _Nullable)error;

- (void)setVideoMirrorMode:(AgoraRteVideoMirrorMode)mode error:(AgoraRteError* _Nullable)error;

- (AgoraRteVideoMirrorMode)videoMirrorMode:(AgoraRteError * _Nullable)error;

- (void)setCropArea:(AgoraRteRect* _Nonnull)cropArea error:(AgoraRteError * _Nullable)error;

- (AgoraRteRect* _Nonnull)cropArea:(AgoraRteError* _Nullable)error;

@end


/**
 * The Canvas class is used to manage the canvas.
 * @since v4.4.0
 * @technical preview
 */
__attribute__((visibility("default"))) @interface AgoraRteCanvas : NSObject

- (instancetype _Nonnull)initWithRte:(AgoraRte* _Nonnull)rte initialConfig:(AgoraRteCanvasInitialConfig * _Nullable)config;

- (BOOL)getConfigs:(AgoraRteCanvasConfig* _Nonnull)config error:(AgoraRteError* _Nullable)error;

- (BOOL)setConfigs:(AgoraRteCanvasConfig* _Nonnull)config error:(AgoraRteError* _Nullable)error;

- (BOOL)addView:(AgoraRteView * _Nonnull)view config:(AgoraRteViewConfig* _Nullable)config error:(AgoraRteError* _Nullable)error;

- (BOOL)removeView:(AgoraRteView * _Nonnull)view config:(AgoraRteViewConfig* _Nullable)config error:(AgoraRteError* _Nullable)error;

@end
